/*
 * Copyright (c) 2001, DecisionSoft Limited All rights reserved.
 * Please see LICENSE.TXT for more information.
 */

#include "ATDateTimeTester.hpp"
#include <pathan/internal/DOMutils/XStr.hpp>
#include <xercesc/validators/schema/SchemaSymbols.hpp>
#include <pathan/XPath2MemoryManager.hpp>
#include <pathan/DynamicContext.hpp>
#include <pathan/internal/factory/DatatypeFactory.hpp>

#include <pathan/ATDateTimeOrDerived.hpp>
#include <pathan/ATDecimalOrDerived.hpp>
#include <pathan/ATDurationOrDerived.hpp>
#include <pathan/Timezone.hpp>

ATDateTimeTester::ATDateTimeTester(XPath2MemoryManager* memMgr) : DatatypeTester(memMgr, "ATDateTimeOrDerived") {
}

void ATDateTimeTester::run(const DynamicContext* context) {
  // test 1 -- equality
  const ATDateTimeOrDerived* dateTime1 = this->createDateTime(X("2000-02-29T13:20:10.0564+02:00"), context);
  const ATDateTimeOrDerived* dateTime2 = this->createDateTime(X("2000-02-29T15:20:10.0564+04:00"), context);

  assertObjectEquals(dateTime1, dateTime1, context);
  assertObjectEquals(dateTime1, dateTime2, context);

  assertEquals(dateTime1->asString(context), X("2000-02-29T11:20:10.0564Z"));
  assertEquals(dateTime1->asString(context), dateTime2->asString(context));
  
  // test 2 -- inequality
  const ATDateTimeOrDerived* dateTime3 = this->createDateTime(X("2002-04-02T06:00:00+01:00"), context);
  assertCondition(!dateTime1->equals(dateTime3, context));
  assertCondition(dateTime1->lessThan(dateTime3, context));
  assertCondition(dateTime3->greaterThan(dateTime1, context));
  
  // test 3 -- component extraction
  assertObjectEquals(dateTime1->getYears(), DatatypeFactory::POD2AT::createInteger(getMemoryManager(), 2000, context), context);
  assertObjectEquals(dateTime1->getMonths(), DatatypeFactory::POD2AT::createInteger(getMemoryManager(), 2, context), context);
  assertObjectEquals(dateTime1->getDays(), DatatypeFactory::POD2AT::createInteger(getMemoryManager(), 29, context), context);
  assertObjectEquals(dateTime1->getHours(), DatatypeFactory::POD2AT::createInteger(getMemoryManager(), 13, context), context);
  assertObjectEquals(dateTime1->getMinutes(), DatatypeFactory::POD2AT::createInteger(getMemoryManager(), 20, context), context);
  assertObjectEquals(dateTime1->getSeconds(), DatatypeFactory::POD2AT::createDecimal(getMemoryManager(), 10.0564, context), context);
  assertCondition(dateTime1->hasTimezone());
  assertCondition(dateTime1->getTimezone()->equals(new (getMemoryManager()) Timezone(2, 0)));

  // test 4 -- adding and subtracting durations/dates
  const ATDurationOrDerived* duration1 = DatatypeFactory::STR2AT::createDayTimeDuration(getMemoryManager(), X("P3DT10H"), context);
  const ATDateTimeOrDerived* result1 = dateTime1->subtractDayTimeDuration(duration1, context);
  assertEquals(result1->asString(context), X("2000-02-26T01:20:10.0564Z"));
  const ATDateTimeOrDerived* result2 = result1->addDayTimeDuration(duration1, context);
  assertObjectEquals(dateTime1, result2, context);

  const ATDurationOrDerived* duration2 = DatatypeFactory::STR2AT::createYearMonthDuration(getMemoryManager(), X("P1Y2M"), context);
  const ATDateTimeOrDerived* result3 = dateTime1->addYearMonthDuration(duration2, context);
  assertEquals(result3->asString(context), X("2001-04-29T11:20:10.0564Z"));
  const ATDateTimeOrDerived* result4 = result3->subtractYearMonthDuration(duration2, context);
  assertObjectEquals(dateTime1, result4, context);
 
  const ATDurationOrDerived* result5 = dateTime3->subtractDateTimeAsDayTimeDuration(dateTime1, context);
  const ATDurationOrDerived* difference = DatatypeFactory::STR2AT::createDayTimeDuration(getMemoryManager(), X("P762DT17H39M49.9436S"), context);
  assertObjectEquals(result5, difference, context);

  // test 5 -- timezones
  const ATDateTimeOrDerived* dateTime4 = this->createDateTime(X("1867-07-01T15:20:00"), context);
  assertCondition(!dateTime4->hasTimezone());

  const ATDateTimeOrDerived* dateTime5 = dateTime4->addTimezone(DatatypeFactory::STR2AT::createDayTimeDuration(getMemoryManager(), X("-PT5H"), context) ,context);
  assertCondition(dateTime5->hasTimezone());
  assertCondition(dateTime5->getTimezone()->equals(new (getMemoryManager()) Timezone(-5, 0)));

  const ATDateTimeOrDerived* dateTime6 = dateTime5->setTimezone(new (getMemoryManager()) Timezone(-8, 0), context);
  assertCondition(dateTime6->getTimezone()->equals(new (getMemoryManager()) Timezone(-8, 0)));

  const ATDateTimeOrDerived* dateTime7 = dateTime6->normalize(context);
  const ATDateTimeOrDerived* dateTime8 = this->createDateTime(X("1867-07-01T23:20:00Z"), context);
  assertObjectEquals(dateTime7, dateTime8, context);

  // test 6 -- typeURI and typeName
  assertEquals(dateTime3->getTypeName(), XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgDT_DATETIME);
  assertEquals(dateTime3->getTypeName(), dateTime3->getPrimitiveTypeName());
  
  assertEquals(dateTime3->getTypeURI(), dateTime3->getPrimitiveTypeURI());
  assertEquals(dateTime3->getTypeURI(), XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgURI_SCHEMAFORSCHEMA);
}

const ATDateTimeOrDerived* ATDateTimeTester::createDateTime(const XMLCh* value,
                                   const DynamicContext* context) {
  return DatatypeFactory::STR2AT::createDateTimeOrDerived(
      getMemoryManager(), XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgURI_SCHEMAFORSCHEMA,
      XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgDT_DATETIME,
      value, context);
}

