/*
 * Copyright (c) 2001, DecisionSoft Limited All rights reserved.
 * Please see LICENSE.TXT for more information.
 */

#include "ATDoubleTester.hpp"
#include <pathan/internal/DOMutils/XStr.hpp>
#include <xercesc/validators/schema/SchemaSymbols.hpp>
#include <pathan/XPath2MemoryManager.hpp>
#include <pathan/internal/m_apm.h>
#include <pathan/DynamicContext.hpp>
#include <pathan/internal/factory/DatatypeFactory.hpp>

#include <pathan/ATDoubleOrDerived.hpp>

ATDoubleTester::ATDoubleTester(XPath2MemoryManager* memMgr) : DatatypeTester(memMgr, "ATDoubleOrDerived") {
}

void ATDoubleTester::run(const DynamicContext* context) {
  // test 1 -- equality
  const ATDoubleOrDerived* double1 = DatatypeFactory::POD2AT::createDouble(getMemoryManager(), 10, context);
  const ATDoubleOrDerived* double2 = DatatypeFactory::POD2AT::createDouble(getMemoryManager(), 10.0, context);
  const ATDoubleOrDerived* double3 = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("10"), context);
  const ATDoubleOrDerived* double4 = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("10.0"), context);
  const ATDoubleOrDerived* double5 = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("1.0E+1"), context);
  const ATDoubleOrDerived* double6 = DatatypeFactory::POD2AT::createDouble(getMemoryManager(), 1.0e1, context);
  const ATDoubleOrDerived* double7 = DatatypeFactory::POD2AT::createDouble(getMemoryManager(), 100e-1, context);

  assertObjectEquals(double1, double2, context);
  assertObjectEquals(double1, double3, context);
  assertObjectEquals(double1, double4, context);
  assertObjectEquals(double1, double5, context);
  assertObjectEquals(double1, double6, context);
  assertObjectEquals(double1, double7, context);
  
  // test 2 -- asString
  assertEquals(double1->asString(context), X("10"));
  assertEquals(double2->asString(context), X("10"));
  assertEquals(double3->asString(context), X("10"));
  assertEquals(double4->asString(context), X("10"));
  assertEquals(double5->asString(context), X("10"));
  assertEquals(double6->asString(context), X("10"));
  assertEquals(double7->asString(context), X("10"));

  // test 3 -- inequality
  const ATDoubleOrDerived* double8 = DatatypeFactory::POD2AT::createDouble(getMemoryManager(), 100, context);
  const ATDoubleOrDerived* double9 = DatatypeFactory::POD2AT::createDouble(getMemoryManager(), -100, context);
  const ATDoubleOrDerived* double10 = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("0.2e+1"), context);
  const ATDoubleOrDerived* double11 = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("0.5e+1"), context);
  const ATDoubleOrDerived* double12 = DatatypeFactory::POD2AT::createDouble(getMemoryManager(), 2.1479, context);
  const ATDoubleOrDerived* double13 = DatatypeFactory::POD2AT::createDouble(getMemoryManager(), 4.5, context);
  const ATDoubleOrDerived* double14 = DatatypeFactory::POD2AT::createDouble(getMemoryManager(), 1, context);
  
  assertCondition(!double1->equals(double8, context));
  assertCondition(!double1->equals(double9, context));
  assertCondition(!double1->equals(double10, context));

  assertCondition(double1->lessThan(double8, context));
  assertCondition(double1->greaterThan(double9, context));
  assertCondition(double1->greaterThan(double10, context));

  // test 4 -- special values (NaN, INF, etc.)
  const ATDoubleOrDerived* doubleNaN = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("NaN"), context);
  const ATDoubleOrDerived* doubleINF = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("INF"), context);
  const ATDoubleOrDerived* doubleNegINF = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("-INF"), context);
  const ATDoubleOrDerived* doubleZero = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("0"), context);
  const ATDoubleOrDerived* doubleNegZero = DatatypeFactory::STR2AT::createDouble(getMemoryManager(), X("-0"), context);

  
  assertCondition(doubleNaN->isNaN());
  assertEquals(doubleNaN->asString(context), X("NaN"));
  
  assertCondition(doubleINF->isInfinite());
  assertCondition(doubleINF->isPositive());
  assertCondition(!doubleINF->isNegative());
  assertEquals(doubleINF->asString(context), X("INF"));

  
  assertCondition(doubleNegINF->isInfinite());
  assertCondition(!doubleNegINF->isPositive());
  assertCondition(doubleNegINF->isNegative());
  assertEquals(doubleNegINF->asString(context), X("-INF"));
  
  assertCondition(doubleZero->isZero());
  assertCondition(doubleZero->isPositive());
  assertCondition(!doubleZero->isNegative());
  assertEquals(doubleZero->asString(context), X("0"));
  
  
  assertCondition(doubleNegZero->isZero());
  assertCondition(!doubleNegZero->isPositive());
  assertCondition(doubleNegZero->isNegative());
  assertEquals(doubleNegZero->asString(context), X("-0"));

  // test 5 -- operators (add, subtract, divide, etc.)
  const Numeric* result1 = double1->add(double1, context);
  const Numeric* result2 = double1->subtract(double1, context);
  const Numeric* result3 = double1->multiply(double10, context);
  const Numeric* result4 = double8->divide(double1, context);
  const Numeric* result5 = double11->mod(double10, context);
  const Numeric* result6 = double12->floor(context);
  const Numeric* result7 = double13->round(context);

  assertObjectEquals(result1, result3, context);
  assertObjectEquals(result2, doubleZero, context);
  assertObjectEquals(result4, double1, context);
  assertObjectEquals(result5, double14, context);
  assertObjectEquals(result6, double10, context);
  assertObjectEquals(result7, double11, context);
  
  //add -- special values
  const ATDoubleOrDerived* result8 = static_cast<const ATDoubleOrDerived*>(double1->add(doubleNaN, context));
  const ATDoubleOrDerived* result9 = static_cast<const ATDoubleOrDerived*>(double1->add(doubleINF, context));
  const ATDoubleOrDerived* result10 = static_cast<const ATDoubleOrDerived*>(double1->add(doubleNegINF, context));
  const Numeric* result11 = double1->add(doubleNegZero, context);
  assertEquals(result8->asString(context), doubleNaN->asString(context));
  assertObjectEquals(result9, doubleINF, context);
  assertObjectEquals(result10, doubleNegINF, context);
  assertObjectEquals(result11, double1, context);

  //subtract -- special values
  const ATDoubleOrDerived* result12 = static_cast<const ATDoubleOrDerived*>(double1->subtract(doubleNaN, context));
  const ATDoubleOrDerived* result13 = static_cast<const ATDoubleOrDerived*>(double1->subtract(doubleINF, context));
  const ATDoubleOrDerived* result14 = static_cast<const ATDoubleOrDerived*>(double1->subtract(doubleNegINF, context));
  const Numeric* result15 = double1->subtract(doubleNegZero, context);
  assertEquals(result12->asString(context), doubleNaN->asString(context));
  assertObjectEquals(result13, doubleNegINF, context);
  assertObjectEquals(result14, doubleINF, context);
  assertObjectEquals(result15, double1, context);
  
  //multiply -- special values
  const ATDoubleOrDerived* result16 = static_cast<const ATDoubleOrDerived*>(double1->multiply(doubleNaN, context));
  const ATDoubleOrDerived* result17 = static_cast<const ATDoubleOrDerived*>(double1->multiply(doubleINF, context));
  const ATDoubleOrDerived* result18 = static_cast<const ATDoubleOrDerived*>(double1->multiply(doubleNegINF, context));
  const Numeric* result19 = double1->multiply(doubleNegZero, context);
  assertEquals(result16->asString(context), doubleNaN->asString(context));
  assertObjectEquals(result17, doubleINF, context);
  assertObjectEquals(result18, doubleNegINF, context);
  assertObjectEquals(result19, doubleNegZero, context);

  //divide -- special values
  const ATDoubleOrDerived* result20 = static_cast<const ATDoubleOrDerived*>(double1->divide(doubleNaN, context));
  const ATDoubleOrDerived* result21 = static_cast<const ATDoubleOrDerived*>(double1->divide(doubleINF, context));
  const ATDoubleOrDerived* result22 = static_cast<const ATDoubleOrDerived*>(double1->divide(doubleNegINF, context));
  const ATDoubleOrDerived* result23 = static_cast<const ATDoubleOrDerived*>(double1->divide(doubleNegZero, context));
  assertEquals(result20->asString(context), doubleNaN->asString(context));
  assertObjectEquals(result21, doubleZero, context);
  assertObjectEquals(result22, doubleNegZero, context);
  assertObjectEquals(result23, doubleNegINF, context);
  
  // test type name, type uri
  assertEquals(double1->getTypeName(), XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgDT_DOUBLE);
  assertEquals(double1->getTypeName(), double1->getPrimitiveTypeName());
  
  assertEquals(double1->getTypeURI(), double1->getPrimitiveTypeURI());
  assertEquals(double1->getTypeURI(), XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgURI_SCHEMAFORSCHEMA);
}
