#include "../config/pathan_config.h"
#include <assert.h>
#include <pathan/SequenceType.hpp>

#include <pathan/XPath2Utils.hpp>
#include <pathan/DynamicContext.hpp>
#include <pathan/DocumentCache.hpp>
#include <pathan/XPath2MemoryManager.hpp>
#include <pathan/internal/parser/QName.hpp>
#include <pathan/exceptions/XPath2ErrorException.hpp>
#include "../exceptions/XPath2TypeCastException.hpp"
#include <pathan/exceptions/ContextException.hpp>
#include <pathan/ATStringOrDerived.hpp>
#include <pathan/ATDoubleOrDerived.hpp>
#include <pathan/ATUntypedAtomic.hpp>
#include <pathan/Node.hpp>
#include <pathan/functions/FunctionConstructor.hpp>
#include <pathan/internal/factory/DatatypeFactory.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/util/XMLUniDefs.hpp>
#include <xercesc/validators/schema/SchemaAttDef.hpp>
#include <xercesc/validators/schema/SchemaElementDecl.hpp>

/////////////////////////////////////////////
// SchemaType definition

SchemaType::SchemaType()
{
    _rootIsType=false;
    _root=NULL;
    _steps=NULL;
}

SchemaType::SchemaType(QualifiedName* globalType)
{
    _rootIsType=true;
    _root=globalType;
    _steps=NULL;
}

SchemaType::~SchemaType()
{
	delete _root;
	delete _steps;
}

SequenceType::SequenceType(const XMLCh* typeURI,const XMLCh* typeName, SequenceType::OccurrenceIndicator occur /* = EXACTLY_ONE */)
{
  // Possible memory leak? - jpcs
	m_pItemType=new SequenceType::ItemType(ItemType::TEST_ATOMIC_TYPE,NULL,new QualifiedName(XERCES_CPP_NAMESPACE_QUALIFIER XMLUni::fgZeroLenString, typeName));
	m_pItemType->setTypeURI(typeURI);
	m_nOccurrence=occur;
}

SequenceType::SequenceType(SequenceType::ItemType* test, OccurrenceIndicator occur /*= EXACTLY_ONE*/)
{
  m_pItemType=test;
  m_nOccurrence=occur;
}

SequenceType::SequenceType()
{
	m_pItemType=0;
	m_nOccurrence=EXACTLY_ONE;
}

SequenceType::~SequenceType()
{
	if(m_pItemType)
		delete m_pItemType;
}

bool SequenceType::matches(const Sequence& toBeTested, DynamicContext* context) const
{
  // "SequenceType matching between a given value and a given SequenceType is performed as follows:
  //  If the SequenceType is empty, the match succeeds only if the value is an empty sequence."
  if(m_pItemType==NULL)
    return toBeTested.getLength()==0;

  // "If the SequenceType is an ItemType with no OccurrenceIndicator, the match succeeds only if 
  //  the value contains precisely one item and that item matches the ItemType "
  if(m_nOccurrence==EXACTLY_ONE && toBeTested.getLength()!=1)
    return false;

  // "If the SequenceType contains an ItemType and an OccurrenceIndicator, the match succeeds only if 
  //  the number of items in the value matches the OccurrenceIndicator and each of these items matches the ItemType. "
  if(m_nOccurrence==PLUS && toBeTested.getLength()==0)
    return false;

  if(m_nOccurrence==QUESTION_MARK && toBeTested.getLength()>1)
    return false;

  // Now test that each item matches the ItemType
  Sequence::const_iterator end = toBeTested.end();
  for(Sequence::const_iterator i = toBeTested.begin(); i != end; ++i) {
    if(!m_pItemType->matches(*i, context))
      return false;
  }
  return true;
}

void SequenceType::setItemType(SequenceType::ItemType* itemType)
{
  m_pItemType=itemType;
}

SequenceType::ItemType::ItemTestType SequenceType::getItemTestType(void) const {

  return m_pItemType->getItemTestType();
}

SequenceType::OccurrenceIndicator SequenceType::getOccurrenceIndicator(void) const {

  return m_nOccurrence;
}

void SequenceType::setOccurrence(SequenceType::OccurrenceIndicator nOccurrence)
{
  m_nOccurrence=nOccurrence;
}

void SequenceType::setNameURI(const XMLCh* const nameURI) {

  if(m_pItemType!=NULL) {
    m_pItemType->setNameURI(nameURI);
  }
}

const XMLCh* SequenceType::getNameURI(DynamicContext* context) const {

  if(m_pItemType!=NULL) {
    return m_pItemType->getNameURI(context);
  }
  return 0;
}

void SequenceType::setTypeURI(const XMLCh* const typeURI) {

  //cerr << "SequnceType:: Setting typeURI: " << typeURI << endl;

  if(m_pItemType!=NULL) {
    m_pItemType->setTypeURI(typeURI);
  }
}

const XMLCh* SequenceType::getTypeURI(DynamicContext* context) const {

  if(m_pItemType!=NULL) {
    return m_pItemType->getTypeURI(context);
  }
  return 0;
}

QualifiedName *SequenceType::getConstrainingName(void) const {

  return m_pItemType->getName();
}

QualifiedName *SequenceType::getConstrainingType(void) const {

  return m_pItemType->getType();
}

SequenceType::ItemType::ItemType(ItemTestType test,QualifiedName* name /*=NULL*/, QualifiedName* type /*=NULL*/)
{
  m_nTestType=test;
  m_pName=name;
  m_pType=type;
  m_pSchemaType=NULL;
  m_TypeURI=m_NameURI=NULL;
  m_bAllowNil=false;
}

SequenceType::ItemType::ItemType(ItemTestType test, QualifiedName* name, SchemaType* type)
{
  m_nTestType=test;
  m_pName=name;
  m_pType=NULL;
  m_pSchemaType=type;
  m_TypeURI=m_NameURI=NULL;
  m_bAllowNil=false;
}

SequenceType::ItemType::~ItemType()
{
    delete m_pName;
    delete m_pType;
	delete m_pSchemaType;
}

SequenceType::ItemType::ItemTestType SequenceType::ItemType::getItemTestType(void) const
{
  return m_nTestType;
}

void SequenceType::ItemType::setAllowNilled(bool value)
{
  m_bAllowNil=value;
}

bool SequenceType::ItemType::getAllowNilled() const
{
  return m_bAllowNil;
}

void SequenceType::ItemType::setName(QualifiedName * name)
{
    m_pName=name;
}

QualifiedName *SequenceType::ItemType::getName(void) const {

  return m_pName;
}

void SequenceType::ItemType::setType(QualifiedName * type)
{
    m_pType=type;
}

QualifiedName *SequenceType::ItemType::getType(void) const {

  return m_pType;
}

void SequenceType::ItemType::setNameURI(const XMLCh* const nameURI) {

  m_NameURI = nameURI;
}

void SequenceType::ItemType::setTypeURI(const XMLCh* const typeURI) {

  m_TypeURI = typeURI;
}

const XMLCh* SequenceType::ItemType::getTypeURI(DynamicContext* context) const
{
	if(m_TypeURI!=0 && *m_TypeURI!=0) {
		//      cerr << "NameTypeConstrain: Returning m_TypeURI: " << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(m_TypeURI) << endl;
		return m_TypeURI;
	}
	//    cerr << "NameTypeConstrain: Returning getUriBoundToPrefix: " << " prefix is: " << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(m_pType->getPrefix()) << " result is: " << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(context->getUriBoundToPrefix(m_pType->getPrefix())) <<  " default uri is " << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(context->getDefaultFuncNS()) << endl;
  if(m_pType != 0)
    return context->getUriBoundToPrefix(m_pType->getPrefix());
  return 0;
}

const XMLCh* SequenceType::ItemType::getNameURI(DynamicContext* context) const
{
	if(m_NameURI!=0 && *m_NameURI!=0)
		return m_NameURI;
  if(m_pName != 0)
    return context->getUriBoundToPrefix(m_pName->getPrefix());
  return 0;
}

bool SequenceType::ItemType::matchesNameType(const Item* toBeTested, DynamicContext* context) const
{
  // Check name constraint
  if(m_pName!=NULL) {
    if(toBeTested->isNode()) {
      //Match resolved prefixs
      const XERCES_CPP_NAMESPACE_QUALIFIER DOMNode* node=((const Node*)toBeTested)->getDOMNode();

      if(!(XPath2Utils::equals(getNameURI(context), node->getNamespaceURI())))
        return false;
      //Match node name
      const XMLCh* nodeName = m_pName->getName();

      if(!(XPath2Utils::equals(nodeName, node->getLocalName())))
        return false;
    }
  }

  //A named atomic type matches a value if the dynamic type of the
  //value is the same as the named atomic type or is derived from the
  //named atomic type by restriction. For example, the ItemType
  //xs:decimal matches the value 12.34 (a decimal literal); it also
  //matches a value whose dynamic type is shoesize, if shoesize is an
  //atomic type derived from xs:decimal.

  if(m_pType) {
    bool result;
    if(toBeTested->isAtomicValue()) {
      result = ((const AnyAtomicType*)toBeTested)->isInstanceOfType(getTypeURI(context), m_pType->getName(), context);
    } else if (toBeTested->isNode()) {
      result = ((const Node*)toBeTested)->hasInstanceOfType(getTypeURI(context), m_pType->getName(), context);
    } else {
      assert(false); // should never get here
    }
    //cerr << "was toBeTested ("<<XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(((const AnyAtomicType*)toBeTested)->getTypeURI()) << ":"
    //     << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(((const AnyAtomicType*)toBeTested)->getTypeName())
    //     << ") instance of "<< XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(getTypeURI(context)) << ":" 
    //     << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(m_pType->getName()) << "? "<< result<<endl;
    
    return result;
  }

  if(m_pSchemaType) {
	bool result=false;
	assert(toBeTested->isNode());
	const XERCES_CPP_NAMESPACE_QUALIFIER DOMNode* pNodeToBeTested=((const Node*)toBeTested)->getDOMNode();
	const DocumentCache* pDocCache=context->getDocumentCache();
	ValidationContext valCtx;
	valCtx.setContextType(ValidationContext::GLOBAL_CONTEXT);
	valCtx=pDocCache->updateValidationContext(valCtx,m_pSchemaType->_root,!m_pSchemaType->_rootIsType,context);
	for(VectorOfQualifiedNames::iterator it=m_pSchemaType->_steps->begin();it!=m_pSchemaType->_steps->end();it++)
		valCtx=pDocCache->updateValidationContext(valCtx,*it,true,context);
	const XMLCh* finalTypeUri=NULL, *finalTypeName=NULL;
	if(pNodeToBeTested->getNodeType()==XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::ATTRIBUTE_NODE)
	{
		valCtx=pDocCache->updateValidationContext(valCtx,m_pName,false,context);
		XERCES_CPP_NAMESPACE_QUALIFIER SchemaAttDef* pAttr=valCtx.getCurrentAttribute();
		finalTypeUri=pAttr->getTypeUri();
		finalTypeName=pAttr->getTypeName();
	}
	else
	{
		valCtx=pDocCache->updateValidationContext(valCtx,m_pName,true,context);
		XERCES_CPP_NAMESPACE_QUALIFIER SchemaElementDecl* pCurEl=valCtx.getCurrentElement();
		finalTypeUri=pCurEl->getTypeUri();
		finalTypeName=pCurEl->getTypeName();
	}

/*	std::cerr << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(finalTypeUri) 
			  << "::" 
			  << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(finalTypeName)
			  << std::endl;
*/
	result = ((const Node*)toBeTested)->hasInstanceOfType(finalTypeUri, finalTypeName, context);
	return result;
  }
  return true;
}

bool SequenceType::ItemType::matches(const XERCES_CPP_NAMESPACE_QUALIFIER DOMNode* toBeTested, DynamicContext* context) const
{
  switch(m_nTestType) {
    case TEST_ELEMENT:
    {
      if(toBeTested->getNodeType()!=XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::ELEMENT_NODE)
        return false;

      if(!matchesNameType(context->getMemoryManager()->createNode(toBeTested), context))
        return false;

      // if the element has xsi:nil="true", m_bAllowNil MUST be true
      const XMLCh* nillable=((XERCES_CPP_NAMESPACE_QUALIFIER DOMElement*)toBeTested)->getAttributeNS( XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgURI_XSI,
                                                      XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgATT_NILL);
      if(XERCES_CPP_NAMESPACE_QUALIFIER XMLString::equals(nillable,XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgATTVAL_TRUE) &&
        !m_bAllowNil)
        return false;

      return true;
    }//case

    case TEST_ATTRIBUTE:
    {
      if(toBeTested->getNodeType()!=XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::ATTRIBUTE_NODE)
        return false;
      if(!matchesNameType(context->getMemoryManager()->createNode(toBeTested), context))
        return false;
      return true;
    }//case

    case TEST_NODE:
    {
      return true;
    }

    case TEST_PI:
    {
       return (toBeTested->getNodeType()==XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::PROCESSING_INSTRUCTION_NODE);
    }

    case TEST_COMMENT:
    {
      return (toBeTested->getNodeType()==XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::COMMENT_NODE);
    }

    case TEST_TEXT:
    {
      return (toBeTested->getNodeType()==XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::TEXT_NODE || toBeTested->getNodeType()==XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::CDATA_SECTION_NODE);
    }

    case TEST_DOCUMENT:
    {
        if(toBeTested->getNodeType()!=XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::DOCUMENT_NODE)
            return false;
        if(m_pName!=NULL || m_pType!=NULL)
        {
            // if we have a constraint on name/type, they apply to the document element
            XERCES_CPP_NAMESPACE_QUALIFIER DOMElement* docRoot=((XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument*)toBeTested)->getDocumentElement();
            if(docRoot==NULL)
                return false;
            if(!matchesNameType(context->getMemoryManager()->createNode(docRoot),context))
                return false;            
        }
        return true;
    }

    case TEST_ANYTHING:
    {
      return true;
    }

    case TEST_ATOMIC_TYPE:
    {
        return false;
    }//case
  }//switch
  return true;
}

bool SequenceType::ItemType::matches(const Item* toBeTested, DynamicContext* context) const
{
  if(toBeTested->isNode())
    return matches(((const Node*)toBeTested)->getDOMNode(),context);
    
  switch(m_nTestType) {
    case TEST_ELEMENT:  
    case TEST_ATTRIBUTE: 
    case TEST_NODE:
    case TEST_PI:
    case TEST_COMMENT:
    case TEST_TEXT:
    case TEST_DOCUMENT:
    {
      return false;
    }
    
    case TEST_ANYTHING:
    {
      return true;
    }

    ///this is wrong said nfg and gareth -- we have non atomic types now.  crioux thinks this is correct (we no longer have non-atomic types)

    // "A named atomic type matches a value if the dynamic type of the value is the same as the named atomic type or is derived 
    //  from the named atomic type by restriction. For example, the ItemType xs:decimal matches the value 12.34 (a decimal literal); 
    //  it also matches a value whose dynamic type is shoesize, if shoesize is a user-defined atomic type derived from xs:decimal. "
    case TEST_ATOMIC_TYPE:
    {
        // the constraining type could be one of the special XPath 2 types
        QualifiedName *constrainingType = getType();

        if(constrainingType != NULL)
        {
            const XMLCh *uri = getTypeURI(context);
            if(XERCES_CPP_NAMESPACE_QUALIFIER XMLString::equals(uri,FunctionConstructor::XMLChXPath2DatatypesURI))
            {
                if(XERCES_CPP_NAMESPACE_QUALIFIER XMLString::equals(constrainingType->getName(),AnyAtomicType::fgDT_ANYATOMICTYPE))
                    return true;
                else if(XERCES_CPP_NAMESPACE_QUALIFIER XMLString::equals(constrainingType->getName(),ATUntypedAtomic::fgDT_UNTYPEDATOMIC))
                    return true;
            }
        }
        if(!matchesNameType(toBeTested, context))
            return false;

        return true;
    }//case

  }//switch
  return true;
}

Sequence SequenceType::castAsFunction(const Sequence &param, DynamicContext* context) const {

  // the first step os to preprocess the function arguments, taking into consideration the XPath 1.0 compatibility flag

  Sequence toBeCasted(context->getMemoryManager());

  // If XPath 1.0 compatibility mode is true, then one of the following conversions is applied:
  if(context->getXPath1CompatibilityMode())
  {
    // 1. If the expected type is a string, optional string, or derived from a string or optional string, 
    //    then the given value V is effectively replaced by fn:string(fn:subsequence(V, 1, 1)).
    if(getItemTestType() == ItemType::TEST_ATOMIC_TYPE &&
       (getOccurrenceIndicator() == SequenceType::EXACTLY_ONE || getOccurrenceIndicator() == SequenceType::QUESTION_MARK) &&
        context->isTypeOrDerivedFromType(getTypeURI(context),
                                         getConstrainingType()->getName(),
                                         XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgURI_SCHEMAFORSCHEMA,
                                         XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgDT_STRING ) )
    {
      // note: atomization is specified in the Formal Semantics
      Sequence atomized = param.atomize(context);
      if(!atomized.isEmpty())
      {
        const Item* atom=atomized.first();
        if(atom->isNode())
          toBeCasted.addItem(DatatypeFactory::STR2AT::createString(context->getMemoryManager(), ((const Node*)atom)->dmStringValue(context), context));
        else
          toBeCasted.addItem(DatatypeFactory::STR2AT::createString(context->getMemoryManager(), atom->asString(context), context));
      }
    }
    // 2. If the expected type is xs:double or xs:double?, then the given value 
    //    V is effectively replaced by fn:number(fn:subsequence(V, 1, 1)).
    else if(getItemTestType() == ItemType::TEST_ATOMIC_TYPE &&
      (getOccurrenceIndicator() == SequenceType::EXACTLY_ONE || getOccurrenceIndicator() == SequenceType::QUESTION_MARK) &&
       context->isTypeOrDerivedFromType(getTypeURI(context),getConstrainingType()->getName(),
         XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgURI_SCHEMAFORSCHEMA,XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgDT_DOUBLE))
    {
      // note: atomization is specified in the Formal Semantics
      Sequence atomized = param.atomize(context);
      if(!atomized.isEmpty())
      {
        const Item* item = atomized.first();
        if(item->isNode()) {
          toBeCasted.addItem(DatatypeFactory::STR2AT::createDouble(context->getMemoryManager(), ((const Node*)item)->dmStringValue(context), context));
        } else {
          const AnyAtomicType* atom = (const AnyAtomicType*)item;
          toBeCasted.addItem((const ATDoubleOrDerived*)atom->castAs(XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgURI_SCHEMAFORSCHEMA, XERCES_CPP_NAMESPACE_QUALIFIER SchemaSymbols::fgDT_DOUBLE, context));
        }
      }
    }
    // 3. If the expected type is a (possibly optional) node or item, then the given value V is effectively 
    //    replaced by fn:subsequence(V, 1, 1).
    else if((getItemTestType() == ItemType::TEST_NODE || getItemTestType() == ItemType::TEST_ANYTHING) && 
          (getOccurrenceIndicator() == SequenceType::EXACTLY_ONE || getOccurrenceIndicator() == SequenceType::QUESTION_MARK)
         )
    {
      // fn:subsequence(expr, 1, 1)
      if(!param.isEmpty())
        toBeCasted.addItem(param.first());
    }
    // 4. Otherwise, the given value is unchanged
    else
      toBeCasted=param;
      
  }
  // If the expected type is a sequence of an atomic type (possibly with an occurrence indicator *, +, or ?), 
  // the following conversions are applied:
  else if(getItemTestType() == ItemType::TEST_ATOMIC_TYPE) 
  {
    QualifiedName *constrainingType = getConstrainingType();

    if(getConstrainingName() != NULL || constrainingType == NULL)
      assert(false);

    const XMLCh* typeURI = getTypeURI(context);
    const XMLCh* typeName = constrainingType->getName();
    //cerr << "typeURI: " << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(typeURI) << endl;
    //cerr << "typeName: " << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(typeName) << endl;

    // 1. Atomization is applied to the given value, resulting in a sequence of atomic values.
    Sequence atomized=param.atomize(context);
    for(Sequence::const_iterator i = atomized.begin(); i != atomized.end(); i++)
    {
      // 2. Each item in the atomic sequence that is of type xdt:untypedAtomic is cast to the required atomic type.
      // crioux thinks this should also add: unless the target type is anyAtomicType!
      if((*i)->isAtomicValue() && ((const AnyAtomicType*)*i)->getPrimitiveTypeIndex() == AnyAtomicType::UNTYPED_ATOMIC &&
          !( XERCES_CPP_NAMESPACE_QUALIFIER XMLString::equals(typeName, AnyAtomicType::fgDT_ANYATOMICTYPE) && 
             XERCES_CPP_NAMESPACE_QUALIFIER XMLString::equals(typeURI, FunctionConstructor::XMLChXPath2DatatypesURI)) )
      {
        try {
          toBeCasted.addItem( ((const AnyAtomicType*)*i)->castAs(typeURI, typeName, context) );
        }
        catch (XPath2TypeCastException &e) {
          DSLthrow(XPath2ErrorException, X("SequenceType::castAsFunction()"), X("Casting from xdt:untypedAtomic to required type failed."));
        } catch (const XERCES_CPP_NAMESPACE_QUALIFIER XMLException& e) {
          DSLthrow(XPath2ErrorException, X("SequenceType::castAsFunction()"), X("Casting from xdt:untypedAtomic to required type failed."));
        }
      }
      // 3. For each numeric item in the atomic sequence that can be promoted to the expected atomic type 
      //    using the promotion rules in B.1 Type Promotion, the promotion is done.
      else if((*i)->isAtomicValue() && ((const AnyAtomicType*)*i)->isNumericValue())
      {
        try {
          const Numeric* promotedType = ((const Numeric*)*i)->promoteTypeIfApplicable(typeURI, typeName, context);
          if(promotedType)
            toBeCasted.addItem( promotedType );
          else
            toBeCasted.addItem( *i );
        } catch (XPath2TypeCastException &e) {
          DSLthrow(XPath2ErrorException, X("SequenceType::castAsFunction()"), X("Type promotion failed (for promotable type)"));
        } catch (const XERCES_CPP_NAMESPACE_QUALIFIER XMLException& e) {
          DSLthrow(XPath2ErrorException, X("SequenceType::castAsFunction()"), X("Type promotion failed (for promotable type)"));
        }


      } else {
        toBeCasted.addItem( *i );
      }
    }
  }
  else 
    toBeCasted=param;

  // If, after the above conversions, the resulting value does not match the expected type according 
  // to the rules for SequenceType Matching, a type error is raised
  if(matches(toBeCasted, context))
    return toBeCasted;
  DSLthrow(XPath2ErrorException,X("SequenceType::castAsFunction()"), X("SequenceType matching failed."));
}

const SequenceType::ItemType *SequenceType::getItemType() const {
  return m_pItemType;
}
